"""
Import all essential functions and constants to re-export them here for easy
access.

This module contains also various pre-defined ISO 8601 format strings.
"""

from isodate.duration import Duration
from isodate.isodates import date_isoformat, parse_date
from isodate.isodatetime import datetime_isoformat, parse_datetime
from isodate.isoduration import duration_isoformat, parse_duration
from isodate.isoerror import ISO8601Error
from isodate.isostrf import (
    D_ALT_BAS,
    D_ALT_BAS_ORD,
    D_ALT_EXT,
    D_ALT_EXT_ORD,
    D_DEFAULT,
    D_WEEK,
    DATE_BAS_COMPLETE,
    DATE_BAS_MONTH,
    DATE_BAS_ORD_COMPLETE,
    DATE_BAS_WEEK,
    DATE_BAS_WEEK_COMPLETE,
    DATE_CENTURY,
    DATE_EXT_COMPLETE,
    DATE_EXT_MONTH,
    DATE_EXT_ORD_COMPLETE,
    DATE_EXT_WEEK,
    DATE_EXT_WEEK_COMPLETE,
    DATE_YEAR,
    DT_BAS_COMPLETE,
    DT_BAS_ORD_COMPLETE,
    DT_BAS_WEEK_COMPLETE,
    DT_EXT_COMPLETE,
    DT_EXT_ORD_COMPLETE,
    DT_EXT_WEEK_COMPLETE,
    TIME_BAS_COMPLETE,
    TIME_BAS_MINUTE,
    TIME_EXT_COMPLETE,
    TIME_EXT_MINUTE,
    TIME_HOUR,
    TZ_BAS,
    TZ_EXT,
    TZ_HOUR,
    strftime,
)
from isodate.isotime import parse_time, time_isoformat
from isodate.isotzinfo import parse_tzinfo, tz_isoformat
from isodate.tzinfo import LOCAL, UTC, FixedOffset
from isodate.version import version as __version__

__all__ = [
    "parse_date",
    "date_isoformat",
    "parse_time",
    "time_isoformat",
    "parse_datetime",
    "datetime_isoformat",
    "parse_duration",
    "duration_isoformat",
    "ISO8601Error",
    "parse_tzinfo",
    "tz_isoformat",
    "UTC",
    "FixedOffset",
    "LOCAL",
    "Duration",
    "strftime",
    "DATE_BAS_COMPLETE",
    "DATE_BAS_ORD_COMPLETE",
    "DATE_BAS_WEEK",
    "DATE_BAS_WEEK_COMPLETE",
    "DATE_CENTURY",
    "DATE_EXT_COMPLETE",
    "DATE_EXT_ORD_COMPLETE",
    "DATE_EXT_WEEK",
    "DATE_EXT_WEEK_COMPLETE",
    "DATE_YEAR",
    "DATE_BAS_MONTH",
    "DATE_EXT_MONTH",
    "TIME_BAS_COMPLETE",
    "TIME_BAS_MINUTE",
    "TIME_EXT_COMPLETE",
    "TIME_EXT_MINUTE",
    "TIME_HOUR",
    "TZ_BAS",
    "TZ_EXT",
    "TZ_HOUR",
    "DT_BAS_COMPLETE",
    "DT_EXT_COMPLETE",
    "DT_BAS_ORD_COMPLETE",
    "DT_EXT_ORD_COMPLETE",
    "DT_BAS_WEEK_COMPLETE",
    "DT_EXT_WEEK_COMPLETE",
    "D_DEFAULT",
    "D_WEEK",
    "D_ALT_EXT",
    "D_ALT_BAS",
    "D_ALT_BAS_ORD",
    "D_ALT_EXT_ORD",
    "__version__",
]



"""
This module provides an ISO 8601:2004 time zone info parser.

It offers a function to parse the time zone offset as specified by ISO 8601.
"""

import re

from isodate.isoerror import ISO8601Error
from isodate.tzinfo import UTC, ZERO, FixedOffset

TZ_REGEX = (
    r"(?P<tzname>(Z|(?P<tzsign>[+-])" r"(?P<tzhour>[0-9]{2})(:?(?P<tzmin>[0-9]{2}))?)?)"
)

TZ_RE = re.compile(TZ_REGEX)


def build_tzinfo(tzname, tzsign="+", tzhour=0, tzmin=0):
    """
    create a tzinfo instance according to given parameters.

    tzname:
      'Z'       ... return UTC
      '' | None ... return None
      other     ... return FixedOffset
    """
    if tzname is None or tzname == "":
        return None
    if tzname == "Z":
        return UTC
    tzsign = ((tzsign == "-") and -1) or 1
    return FixedOffset(tzsign * tzhour, tzsign * tzmin, tzname)


def parse_tzinfo(tzstring):
    """
    Parses ISO 8601 time zone designators to tzinfo objects.

    A time zone designator can be in the following format:
              no designator indicates local time zone
      Z       UTC
      +-hhmm  basic hours and minutes
      +-hh:mm extended hours and minutes
      +-hh    hours
    """
    match = TZ_RE.match(tzstring)
    if match:
        groups = match.groupdict()
        return build_tzinfo(
            groups["tzname"],
            groups["tzsign"],
            int(groups["tzhour"] or 0),
            int(groups["tzmin"] or 0),
        )
    raise ISO8601Error("%s not a valid time zone info" % tzstring)


def tz_isoformat(dt, format="%Z"):
    """
    return time zone offset ISO 8601 formatted.
    The various ISO formats can be chosen with the format parameter.

    if tzinfo is None returns ''
    if tzinfo is UTC returns 'Z'
    else the offset is rendered to the given format.
    format:
        %h ... +-HH
        %z ... +-HHMM
        %Z ... +-HH:MM
    """
    tzinfo = dt.tzinfo
    if (tzinfo is None) or (tzinfo.utcoffset(dt) is None):
        return ""
    if tzinfo.utcoffset(dt) == ZERO and tzinfo.dst(dt) == ZERO:
        return "Z"
    tdelta = tzinfo.utcoffset(dt)
    seconds = tdelta.days * 24 * 60 * 60 + tdelta.seconds
    sign = ((seconds < 0) and "-") or "+"
    seconds = abs(seconds)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    if hours > 99:
        raise OverflowError("can not handle differences > 99 hours")
    if format == "%Z":
        return "%s%02d:%02d" % (sign, hours, minutes)
    elif format == "%z":
        return "%s%02d%02d" % (sign, hours, minutes)
    elif format == "%h":
        return "%s%02d" % (sign, hours)
    raise ValueError('unknown format string "%s"' % format)



"""
This module provides some datetime.tzinfo implementations.

All those classes are taken from the Python documentation.
"""

import time
from datetime import timedelta, tzinfo

ZERO = timedelta(0)
# constant for zero time offset.


class Utc(tzinfo):
    """UTC

    Universal time coordinated time zone.
    """

    def utcoffset(self, dt):
        """
        Return offset from UTC in minutes east of UTC, which is ZERO for UTC.
        """
        return ZERO

    def tzname(self, dt):
        """
        Return the time zone name corresponding to the datetime object dt,
        as a string.
        """
        return "UTC"

    def dst(self, dt):
        """
        Return the daylight saving time (DST) adjustment, in minutes east
        of UTC.
        """
        return ZERO

    def __reduce__(self):
        """
        When unpickling a Utc object, return the default instance below, UTC.
        """
        return _Utc, ()


UTC = Utc()
# the default instance for UTC.


def _Utc():
    """
    Helper function for unpickling a Utc object.
    """
    return UTC


class FixedOffset(tzinfo):
    """
    A class building tzinfo objects for fixed-offset time zones.

    Note that FixedOffset(0, 0, "UTC") or FixedOffset() is a different way to
    build a UTC tzinfo object.
    """

    def __init__(self, offset_hours=0, offset_minutes=0, name="UTC"):
        """
        Initialise an instance with time offset and name.
        The time offset should be positive for time zones east of UTC
        and negate for time zones west of UTC.
        """
        self.__offset = timedelta(hours=offset_hours, minutes=offset_minutes)
        self.__name = name

    def utcoffset(self, dt):
        """
        Return offset from UTC in minutes of UTC.
        """
        return self.__offset

    def tzname(self, dt):
        """
        Return the time zone name corresponding to the datetime object dt, as a
        string.
        """
        return self.__name

    def dst(self, dt):
        """
        Return the daylight saving time (DST) adjustment, in minutes east of
        UTC.
        """
        return ZERO

    def __repr__(self):
        """
        Return nicely formatted repr string.
        """
        return "<FixedOffset %r>" % self.__name


STDOFFSET = timedelta(seconds=-time.timezone)
# locale time zone offset

# calculate local daylight saving offset if any.
if time.daylight:
    DSTOFFSET = timedelta(seconds=-time.altzone)
else:
    DSTOFFSET = STDOFFSET

DSTDIFF = DSTOFFSET - STDOFFSET
# difference between local time zone and local DST time zone


class LocalTimezone(tzinfo):
    """
    A class capturing the platform's idea of local time.
    """

    def utcoffset(self, dt):
        """
        Return offset from UTC in minutes of UTC.
        """
        if self._isdst(dt):
            return DSTOFFSET
        else:
            return STDOFFSET

    def dst(self, dt):
        """
        Return daylight saving offset.
        """
        if self._isdst(dt):
            return DSTDIFF
        else:
            return ZERO

    def tzname(self, dt):
        """
        Return the time zone name corresponding to the datetime object dt, as a
        string.
        """
        return time.tzname[self._isdst(dt)]

    def _isdst(self, dt):
        """
        Returns true if DST is active for given datetime object dt.
        """
        tt = (
            dt.year,
            dt.month,
            dt.day,
            dt.hour,
            dt.minute,
            dt.second,
            dt.weekday(),
            0,
            -1,
        )
        stamp = time.mktime(tt)
        tt = time.localtime(stamp)
        return tt.tm_isdst > 0


# the default instance for local time zone.
LOCAL = LocalTimezone()
